#ifndef AMDIS_FIRST_ORDER_ASSEMBLER_H
#define AMDIS_FIRST_ORDER_ASSEMBLER_H

#include "SubAssembler.h"

namespace AMDiS {

  class Q10PsiPhi;
  class Q01PsiPhi;
  class Q1Psi;

  /**
   * \ingroup Assembler
   * 
   * \brief
   * SubAssembler for first order terms.
   */
  class FirstOrderAssembler : public SubAssembler
  {
  public:
    MEMORY_MANAGED(FirstOrderAssembler);

    /** \brief
     * Creates and returns the FirstOrderAssembler for Operator op and
     * the given assembler. If all terms are piecewise constant precalculated 
     * integrals can be used while assembling and the returned 
     * ZeroOrderAssembler is of type Pre0. Otherwise a Quad0 object will
     * be returned.
     */
    static FirstOrderAssembler* getSubAssembler(Operator *op,
						Assembler *assembler,
						Quadrature *quadrat,
						FirstOrderType type,
						bool optimized);
  
    /** \brief
     * Destructor.
     */
    virtual ~FirstOrderAssembler() {};

  protected:
    /** \brief
     * Constructor.
     */
    FirstOrderAssembler(Operator *op,
			Assembler *assembler,
			Quadrature *quadrat,
			bool optimized,
			FirstOrderType type);


  protected:
    /** \brief
     * List of all yet created optimized zero order assemblers for grdPsi.
     */
    static std::vector<SubAssembler*> optimizedSubAssemblersGrdPsi;

    /** \brief
     * List of all yet created standard zero order assemblers for grdPsi.
     */
    static std::vector<SubAssembler*> standardSubAssemblersGrdPsi;

    /** \brief
     * List of all yet created optimized zero order assemblers for grdPhi.
     */
    static std::vector<SubAssembler*> optimizedSubAssemblersGrdPhi;

    /** \brief
     * List of all yet created standard zero order assemblers for grdPhi.
     */
    static std::vector<SubAssembler*> standardSubAssemblersGrdPhi;
  };

  // ============================================================================
  // ===== class Stand10 ========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard first order assembler for grdPsi.
   */
  class Stand10 : public FirstOrderAssembler
  {
  public:
    MEMORY_MANAGED(Stand10);

    /** \brief
     * Constructor.
     */
    Stand10(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/);
  };

  // ============================================================================
  // ===== class Stand10 ========================================================
  // ============================================================================

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Standard first order assembler for grdPhi.
   */
  class Stand01 : public FirstOrderAssembler
  {
  public:
    MEMORY_MANAGED(Stand01);

    /** \brief
     * Constructor.
     */
    Stand01(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector *) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Quad10 =========================================================
  // ============================================================================

  /** \brief
   * First order assembler for grdPsi using fast quadratures.
   */
  class Quad10 : public FirstOrderAssembler
  {
  public:
    MEMORY_MANAGED(Quad10);

    /** \brief
     * Constructor.
     */
    Quad10(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/);
  };

  // ============================================================================
  // ===== class Quad01 =========================================================
  // ============================================================================

  /** \brief
   * First order assembler for grdPhi using fast quadratures.
   */
  class Quad01 : public FirstOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Quad01);

    /** \brief
     * Constructor.
     */
    Quad01(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };
  };

  // ============================================================================
  // ===== class Pre10 ==========================================================
  // ============================================================================

  /** \brief
   * First order assembler for grdPsi using precalculated integrals
   */
  class Pre10 : public FirstOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Pre10);

    /** \brief
     * Constructor.
     */
    Pre10(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/);

  protected:
    /** \brief
     * Integral of the product of the derivative of psi and phi.
     */
    const Q10PsiPhi *q10;

    /** \brief
     * Integral of the derivative of psi.
     */
    const Q1Psi *q1;

    friend class FirstOrderAssembler;
  };

  // ============================================================================
  // ===== class Pre01 ==========================================================
  // ============================================================================

  /** \brief
   *  First order assembler for grdPhi using precalculated integrals
   */
  class Pre01 : public FirstOrderAssembler 
  {
  public:
    MEMORY_MANAGED(Pre01);

    /** \brief
     * Constructor.
     */
    Pre01(Operator *op, Assembler *assembler, Quadrature *quad);

    /** \brief
     * Implements SubAssembler::calculateElementMatrix().
     */
    void calculateElementMatrix(const ElInfo *elInfo, ElementMatrix *mat);

    void calculateElementMatrix(const ElInfo *rowElInfo,
				const ElInfo *colElInfo,
				const ElInfo *smallElInfo,
				const ElInfo *largeElInfo,
				ElementMatrix *mat) 
    {
      ERROR_EXIT("CEM not yet\n");
    }

    /** \brief
     * Implements SubAssembler::calculateElementVector().
     */
    void calculateElementVector(const ElInfo *, ElementVector */*vec*/) {
      ERROR_EXIT("should not be called\n");
    };

  protected:
    /** \brief
     * Integral of the product of psi and the derivative of phi.
     */
    const Q01PsiPhi *q01;

    /** \brief
     * Integral of the derivative of phi.
     */
    const Q1Psi *q1;

    friend class FirstOrderAssembler;
  };


}

#endif // AMDIS_FIRST_ORDER_ASSEMBLER_H
