// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file OEMSolver.h */

/**
 * \defgroup Solver Solver module
 * @{ <img src="solver.png"> @}
 * 
 * \brief
 * Contains all classes needed for solving linear and non linear equation
 * systems.
 */

#ifndef AMDIS_OEM_SOLVER_H
#define AMDIS_OEM_SOLVER_H

// ============================================================================
// ===== includes =============================================================
// ============================================================================

#include "Global.h"

namespace AMDiS {

  // ============================================================================
  // ===== forward declarations =================================================
  // ============================================================================

  template<typename VectorType> class Preconditioner;
  template<typename VectorType> class CreatorInterface;
  template<typename VectorType> class MatVecMultiplier;

  // ============================================================================
  // ===== definitions ==========================================================
  // ============================================================================

  /** \brief prints information before solving the system */
#define START_INFO() this->start_info(funcName)

  /** \brief prints information after solving the system */
#define BREAK_INFO(st,it,res,ores)		\
  this->break_info(funcName,st,it,res,ores)

  /** \brief prints information while solving the system */
#define SOLVE_INFO(it,res,ores)			\
  this->solve_info(funcName,it,res,ores)


  // ============================================================================
  // ===== class OEMSolver ======================================================
  // ============================================================================

  /**
   * \ingroup Solver
   * 
   *\brief
   * Solver for linear equation systems.
   * Virtual base class for OEM solvers. 
   */
  template<typename VectorType>
  class OEMSolver
  {
  public:
    /** \brief
     * The constructor reads needed parameters and sets solvers \ref name.
     */
    OEMSolver(std::string name);
  
    /** \brief
     * Virtual destructor because OEMSolver is pure virtual. 
     */
    virtual ~OEMSolver();

    void initParameters();

    /** \brief
     * Solves the linear system.
     */
    int solve(MatVecMultiplier<VectorType> *mv, 
	      VectorType *x, VectorType *b,
	      Preconditioner<VectorType> *lPrecon = NULL, 
	      Preconditioner<VectorType> *rPrecon = NULL,
	      bool reuseMatrix = false)
    {
      FUNCNAME("OEMSolver::solve()");
      TEST_EXIT(mv)("no matVec\n");
      TEST_EXIT(vectorCreator)("no vectorCreator\n");

      leftPrecon = lPrecon;
      rightPrecon = rPrecon;

      init();

      if (leftPrecon) 
	leftPrecon->init();
      
      if (rightPrecon) 
	rightPrecon->init();

      int iter = solveSystem(mv, x, b, reuseMatrix);

      if (leftPrecon) 
	leftPrecon->exit();

      if (rightPrecon) 
	rightPrecon->exit();

      exit();

      return iter;
    }

    // ===== getting-methods ======================================================

    /** \name getting methods
     * \{ 
     */

    /** \brief
     * Returns solvers \ref name.
     */
    inline const std::string& getName() { 
      return name; 
    }

    /** \brief
     * Returns \ref tolerance
     */
    inline double getTolerance() {
      return tolerance;
    }  

    /** \brief
     * Returns \ref max_iter
     */
    inline int getMaxIterations() {
      return max_iter;
    }

    /** \brief
     * Returns \ref residual
     */
    inline double getResidual() {
      return residual;
    }

    /** \} */ 

    // ===== setting-methods ======================================================

    /** \name setting methods
     * \{ 
     */

    /** \brief
     * Sets \ref tolerance
     */
    inline void setTolerance(double tol) {
      tolerance = tol;
    }

    /** \brief
     * Sets \ref relative
     */
    inline void setRelative(bool rel) {
      relative = rel;
    }

    /** \brief
     * Sets \ref max_iter
     */
    inline void setMaxIterations(int i) {
      max_iter = i;
    }

    /** \brief
     * Sets \ref info
     */
    inline void setInfo(int i) {
      info = i;
    }
  
    inline void setVectorCreator(CreatorInterface<VectorType> *creator) {
      vectorCreator = creator;
    }

    inline CreatorInterface<VectorType>* getVectorCreator() {
      return vectorCreator;
    }

    /** \} */

  protected:
    /** \brief
     * To be overloaded by concrete solvers.
     */
    virtual void init() = 0;

    /** \brief
     * To be overloaded by concrete solvers.
     */
    virtual void exit() = 0;

    /** \brief
     * To be overloaded by concrete solvers.
     */
    virtual int solveSystem(MatVecMultiplier<VectorType> *mv,
			    VectorType *x, 
			    VectorType *b,
			    bool reuseMatrix) = 0;


    /** \brief
     * Prints start information. Uses \ref START_INFO.
     */
    void start_info(const char *);

    /** \brief
     * Prints break information. Uses \ref BREAK_INFO.
     */
    void break_info(const char *, const char *, int, double, double *);

    /** \brief
     * Prints solve information. Uses \ref SOLVE_INFO.
     */
    int solve_info(const char *, int, double, double *);

  protected:
    /** \brief
     * solvers name.
     */
    std::string name;

    /** \brief
     * solver tolerance. Set in OEMSolver's constructor. 
     */
    double tolerance;

    /** \brief
     * solver relative. Set in OEMSolver's constructor. 
     */
    bool relative;

    /** \brief
     * maximal number of iterations. Set in OEMSolver's constructor.
     */
    int max_iter;

    /** \brief
     * info level during solving the system. Set in OEMSolver's constructor.
     */
    int info;

    /** \brief
     * current residual.
     */
    double residual;

    Preconditioner<VectorType> *leftPrecon;

    Preconditioner<VectorType> *rightPrecon;

    CreatorInterface<VectorType> *vectorCreator;

  };

  // ============================================================================
  // ===== class OEMSolverCreator ===============================================
  // ============================================================================

  /** 
   * \ingroup Solver
   *
   * \brief
   * Interface for creators of concrete OEMSolvers. 
   */
  template<typename T>
  class OEMSolverCreator : public CreatorInterface<OEMSolver<T> >
  { 
  public:
    virtual ~OEMSolverCreator() {}

    /** \brief
     * Sets \ref problem
     */
    void setName(std::string name_) { 
      name = name_; 
    }

  protected:
    /** \brief
     * Pointer to the problem the solver will belong to. Needed as parameter
     * when constructing an OEMSolver
     */
    std::string name;
  };
}

#include "OEMSolver.hh"

#include "Preconditioner.h"
#include "BiCGSolver.h"
#include "BiCGStab_M.h"
#include "CGSolver.h"
#include "GMResSolver.h"
#include "ODirSolver.h"
#include "OResSolver.h"

#endif // AMDIS_OEM_SOLVER_H
