namespace AMDiS {

  template<typename T>
  SolutionDataStorage<T>::SolutionDataStorage(std::string name)
    : maxMemoryUsage(100),
      useDisk(true),
      writeDirectory(""),
      fixedFESpace(false),
      lastPos(-1),
      poped(false),
      memoryUsage(0)
  {
    solutions.empty();
    feSpaces.empty();
    timestamps.empty();
    
    int tmp = 0;
    GET_PARAMETER(0, name + "->memory usage", "%d", &maxMemoryUsage);
    GET_PARAMETER(0, name + "->use disk", "%d", &tmp);
    useDisk = (tmp == 0) ? false : true;
    GET_PARAMETER(0, name + "->directory", &writeDirectory);
  }

  template<typename T>
  SolutionDataStorage<T>::~SolutionDataStorage()
  {
    clear();
  }

  template<typename T>
  void SolutionDataStorage<T>::setFixFESpace(typename SolutionHelper<T>::type feSpace)
  {
    FUNCNAME("SolutionDataStorage::setFixFESpace()");

    TEST_EXIT(solutions.size() == 0)("Cannot set FE Space, if solutions already stored!\n");

    fixedFESpace = true;
    feSpaces.push_back(feSpace);

    addMemoryUsage(feSpace);
  }

  template<typename T>
  void SolutionDataStorage<T>::push(T *solution,
				    double timestamp)
  {
    FUNCNAME("SolutionDataStorage<T>::push()");

    if (fixedFESpace) {
      ERROR_EXIT("Not yet\n");
    }

    // If pop was the last operation, cleanup and reset the data storage.
    if (poped) {
      clear();
      poped = false;
    }
    
    std::vector<FiniteElemSpace*> feSpace(solution->getFESpaces().size());
    for (int i = 0; i < feSpace.size(); i++) {
      feSpace[i] = NEW FiniteElemSpace();
      *(feSpace[i]) = *(solution->getFESpace(i));       	    
      memoryUsage += feSpace[i]->calcMemoryUsage();
    }
    
    feSpaces.push_back(feSpace);
  	  
    SystemVector *vec = NEW SystemVector("tmp", feSpace, solution->getFESpaces().size());
    vec->createNewDOFVectors("tmp");
    vec->setCoarsenOperation(COARSE_INTERPOL);
    vec->interpol(solution, 1.0);
    memoryUsage += vec->calcMemoryUsage();

    solutions.push_back(vec);
    timestamps.push_back(timestamp);

    lastPos++;
  }

  template<typename T>
  bool SolutionDataStorage<T>::pop(T **solution,
				   double *timestamp)
  {
    if (lastPos < 0) {
      return false;
    }

    *solution = solutions[lastPos];
    *timestamp = timestamps[lastPos];

    lastPos--;
    poped = true;

    return true;
  }

  template<typename T>
  void SolutionDataStorage<T>::clear()
  {
    for (int i = 0; i < static_cast<int>(solutions.size()); i++) {
      DELETE solutions[i];
    }

    if (!fixedFESpace) {
      for (int i = 0; i < static_cast<int>(feSpaces.size()); i++) {
	deleteFeSpace(feSpaces[i]);    
      }
    }

    solutions.clear();
    feSpaces.clear();
    timestamps.clear();

    lastPos = -1;
    memoryUsage = 0;
  }

  template<typename T>
  void SolutionDataStorage<T>::deleteFeSpace(FiniteElemSpace *feSpace) 
  {
    if (feSpace) {
      DELETE feSpace->getMesh();         
      DELETE feSpace;
    }
  }

  template<typename T>	  
  void SolutionDataStorage<T>::addContainer(std::string name)
  {
    std::vector<WorldVector<double> > emptyVec;
    containers.insert(std::pair<std::string, std::vector<WorldVector<double> > >(name, emptyVec) );
    containersPos.insert(std::pair<std::string, int>(name, -1));
  }

  template<typename T>	  
  void SolutionDataStorage<T>::push(std::string name, WorldVector<double> value)
  {
    containers[name].push_back(value);
    containersPos[name] = containers[name].size() - 1;
  }

  template<typename T>	  
  void SolutionDataStorage<T>::pop(std::string name, WorldVector<double> &value)
  {
    value = containers[name][containersPos[name]];
    containersPos[name]--;
  }

  template<typename T>
  void SolutionDataStorage<T>::reset(std::string name)
  {
    containersPos[name] = containers[name].size() - 1;
  }

  template<typename T>
  void SolutionDataStorage<T>::clear(std::string name)
  {
    containers[name].clear();
    containersPos[name] = -1;
  }

}
